/**
 * Copyright (c) 2016  Peter Pentchev
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>

#include <stdbool.h>
#include <string.h>

#include <libxml/parser.h>

#define state_errx(msg)	state->error("%s", (msg))
#define seek_errx(msg)	seek->state->error("%s", (msg))
#define errx(code, msg)	state_errx((msg))
#include "flexarr.h"

#include "axpath_common.h"
#include "axpath_libxml.h"
#include "axpath_libxml_internal.h"

struct axpath_libxml_state *
axpath_libxml_state_init(axpath_found_handler found, axpath_missing_handler missing, axpath_error_handler error)
{
	struct axpath_libxml_state * const state = calloc(1, sizeof(*state));
	if (state == NULL)
		return (NULL);
	state->found = found;
	state->missing = missing;
	state->error = error;
	return (state);
}

void
axpath_libxml_add_pattern(struct axpath_libxml_state * const state, const char * const s)
{

	FLEXARR_ALLOC(state->patterns, 1, state->p_count, state->p_alloc);
	struct axpath_libxml_pattern * const p = &state->patterns[state->p_count - 1];
	p->full = strdup(s);

	FLEXARR_INIT(p->frag, p->f_count, p->f_alloc);
	const char *iter = s;
	if (*iter != '/')
		state->error("Only patterns starting with a slash handled for the present");
	iter++;
	while (true) {
		const char * const frag = strchr(iter, '/');
		const size_t len = frag == NULL? strlen(iter): (size_t)(frag - iter);
		if (len == 0)
			state->error("Invalid pattern, two slashes in a row or a trailing slash");
		FLEXARR_ALLOC(p->frag, 1, p->f_count, p->f_alloc);
		p->frag[p->f_count - 1] = xmlCharStrndup(iter, len);
		if (frag == NULL)
			break;
		iter = frag + 1;
	}
}

static void mp_valid_start_element(struct axpath_libxml_seek *seek, const xmlChar *name, const xmlChar **attrs);
static void mp_invalid_start_element(struct axpath_libxml_seek *seek, const xmlChar *name, const xmlChar **attrs);
static void mp_no_start_element(struct axpath_libxml_seek *seek, const xmlChar *name, const xmlChar **attrs);

static void mp_valid_end_element(struct axpath_libxml_seek *seek);
static void mp_invalid_end_element(struct axpath_libxml_seek *seek);
static void mp_found_end_element(struct axpath_libxml_seek *seek);

static void mp_ignore_characters(struct axpath_libxml_seek *seek, const xmlChar *chars, size_t len);
static void mp_store_characters(struct axpath_libxml_seek *seek, const xmlChar *chars, size_t len);

enum parser_type {
	MON_PARSER_VALID,
	MON_PARSER_INVALID,
	MON_PARSER_FOUND,
	MON_PARSER_COUNT,
};

struct axpath_libxml_parser parser[MON_PARSER_COUNT] = {
	{
		.start_element = mp_valid_start_element,
		.end_element = mp_valid_end_element,
		.characters = mp_ignore_characters,
	},
	{
		.start_element = mp_invalid_start_element,
		.end_element = mp_invalid_end_element,
		.characters = mp_ignore_characters,
	},
	{
		.start_element = mp_no_start_element,
		.end_element = mp_found_end_element,
		.characters = mp_store_characters,
	},
};

static void
end_document_missing(const struct axpath_libxml_state * const state __unused)
{
	state->missing();
}

static void
end_document_found(const struct axpath_libxml_state * const state)
{
	char res_buf[2048];
	size_t ofs = 0;
	for (size_t i_s = 0; i_s < state->s_count; i_s++)
	{
		const struct axpath_libxml_pattern * const p = state->seek[i_s].pattern;
		for (size_t i_m = 0; i_m < p->m_count; i_m++)
		{
			if (ofs > 0) {
				res_buf[ofs] = ' ';
				ofs++;
			}
			int n = snprintf(res_buf + ofs, sizeof(res_buf) - ofs,
			    "%s", p->matches[i_m]);
			if (n < 0)
				state->error("Could not build up the result string");
			ofs += n;
		}
	}
	state->found(res_buf);
}

#undef errx
#define errx(code, msg)	seek_errx((msg))
void mp_valid_start_element(struct axpath_libxml_seek * const seek, const xmlChar * const name, const xmlChar ** const attrs __unused)
{
	if (xmlStrcmp(seek->pattern->frag[seek->c_valid], name) != 0) {
		seek->c_invalid = 1;
		seek->parser = &parser[MON_PARSER_INVALID];
		return;
	}

	seek->c_valid++;
	if (seek->c_valid != seek->pattern->f_count)
		return;

	seek->state->end_document = end_document_found;

	struct axpath_libxml_pattern * const pattern = seek->pattern;
	const size_t old_alloc = pattern->m_alloc;
	FLEXARR_ALLOC(pattern->matches, 1, pattern->m_count, pattern->m_alloc);
	const size_t idx = pattern->m_count - 1;
	pattern->matches[idx] = xmlCharStrdup("");
	if (old_alloc != pattern->m_alloc) {
		pattern->m_length = realloc(pattern->m_length, pattern->m_alloc * sizeof(pattern->m_length[0]));
		if (pattern->m_length == NULL)
			seek->state->error("Out of memory");
	}
	pattern->m_length[idx] = 0;

	seek->parser = &parser[MON_PARSER_FOUND];
}
#undef errx
#define errx(code, msg)	state_errx((msg))

void mp_invalid_start_element(struct axpath_libxml_seek * const seek, const xmlChar * const name __unused, const xmlChar ** const attrs __unused)
{
	seek->c_invalid++;
}

void mp_no_start_element(struct axpath_libxml_seek * const seek, const xmlChar * const name, const xmlChar ** const attrs __unused)
{
	seek->state->error("Unexpected XML element '%s' within the found pattern '%s'", name, seek->pattern->full);
}

void mp_valid_end_element(struct axpath_libxml_seek * const seek)
{
	seek->c_valid--;
}

void mp_invalid_end_element(struct axpath_libxml_seek * const seek)
{
	seek->c_invalid--;
	if (seek->c_invalid == 0)
		seek->parser = &parser[MON_PARSER_VALID];
}

void mp_ignore_characters(struct axpath_libxml_seek * const seek __unused, const xmlChar * const chars __unused, const size_t len __unused)
{
}

void mp_store_characters(struct axpath_libxml_seek * const seek, const xmlChar * const chars, const size_t len)
{
	if (len == 0)
		return;
	struct axpath_libxml_pattern * const pattern = seek->pattern;
	const size_t idx = pattern->m_count - 1;
	const size_t clen = pattern->m_length[idx];
	const size_t sz = clen + len + 1;
	char * const nstr = realloc(pattern->matches[idx], sz);
	if (nstr == NULL)
		seek->state->error("Out of memory");
	memcpy(nstr, pattern->matches[idx], clen);
	memcpy(nstr + clen, chars, len);
	nstr[clen + len] = '\0';
	pattern->m_length[idx] += len;
}

void mp_found_end_element(struct axpath_libxml_seek * const seek)
{
	seek->parser = &parser[MON_PARSER_VALID];
	seek->parser->end_element(seek);
}

static void
p_start_document(void * const xstate __unused)
{
	struct axpath_libxml_state * const state = xstate;
	state->end_document = end_document_missing;

	free(state->seek);
	state->seek = calloc(state->p_count, sizeof(struct axpath_libxml_seek));
	state->s_count = state->p_count;
	for (size_t i = 0; i < state->p_count; i++)
		state->seek[i].pattern = &state->patterns[i];

	for (size_t i = 0; i < state->s_count; i++) {
		struct axpath_libxml_seek * const seek = &state->seek[i];
		seek->state = state;
		seek->parser = &parser[MON_PARSER_VALID];
		seek->c_valid = 0;
		struct axpath_libxml_pattern * const pattern = seek->pattern;
		free(pattern->matches);
		FLEXARR_INIT(pattern->matches, pattern->m_count, pattern->m_alloc);
	}
}

static void
p_end_document(void * const xstate)
{
	const struct axpath_libxml_state * const state = xstate;
	state->end_document(state);
}

static void
p_start_element(void * const xstate, const xmlChar * const name, const xmlChar ** const attrs)
{
	const struct axpath_libxml_state * const state = xstate;

	for (size_t i = 0; i < state->s_count; i++)
		state->seek[i].parser->start_element(&state->seek[i], name, attrs);
}

static void
p_end_element(void * const xstate, const xmlChar * const name __unused)
{
	const struct axpath_libxml_state * const state = xstate;

	for (size_t i = 0; i < state->s_count; i++)
		state->seek[i].parser->end_element(&state->seek[i]);
}

static void
p_characters(void * const xstate, const xmlChar * const chars, const int len)
{
	const struct axpath_libxml_state * const state = xstate;

	for (size_t i = 0; i < state->s_count; i++)
		state->seek[i].parser->characters(&state->seek[i], chars, len);
}

static xmlEntityPtr
p_get_entity(void * const state __unused, const xmlChar * const name)
{
	return xmlGetPredefinedEntity(name);
}

static void
p_warning(void * const state __unused, const char * const fmt __unused, ...)
{
	/* Not sure, just ignore them for the moment */
}

static void
p_error(void * const xstate, const char * const fmt, ...)
{
	const struct axpath_libxml_state * const state = xstate;
	static char s[2048];
	va_list v;
	va_start(v, fmt);
	const int n = vsnprintf(s, sizeof(s), fmt, v);
	if (n < 0)
		state->error("Could not display an XML parser error message");
	for (int i = 0; i < n; i++)
		if (s[i] == '\r' || s[i] == '\n')
			s[i] = ' ';
	state->error("Malformed XML document: %s", s);
}

void
axpath_libxml_parse_chunk(struct axpath_libxml_state * const state, const char * const buf, const size_t sz)
{
	xmlParseChunk(state->parser_context, buf, sz, 0);
}

void
axpath_libxml_parse_end(struct axpath_libxml_state * const state)
{
	xmlParseChunk(state->parser_context, NULL, 0, 1);
}

static xmlSAXHandler axpath_sax = {
	.getEntity = p_get_entity,
	.startDocument = p_start_document,
	.endDocument = p_end_document,
	.startElement = p_start_element,
	.endElement = p_end_element,
	.characters = p_characters,
	.warning = p_warning,
	.error = p_error,
	.fatalError = p_error,
};

void
axpath_libxml_create_parser(struct axpath_libxml_state * const state, const char * const url, const char * const buf, const size_t len)
{
	if (len >= 4) {
		state->parser_context = xmlCreatePushParserCtxt(&axpath_sax, state,
		    buf, len, url);
		if (state->parser_context == NULL)
			state->error("Could not create an XML parser context");
	} else {
		state->parser_context = xmlCreatePushParserCtxt(&axpath_sax, state,
		    NULL, 0, url);
		if (state->parser_context == NULL)
			state->error("Could not create an XML parser context");
		if (len > 0)
			axpath_libxml_parse_chunk(state, buf, len);
	}
}
