/*-
 * Copyright (c) 2016  Peter Pentchev
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/wait.h>

#include <err.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>

#include "flexarr.h"
#include "axpath.h"

struct mon_state {
	const char *url;
	CURL *curlh;
	struct axpath_state *axstate;
	void (*complete)(struct mon_state *);
};

static bool		verbose;

static void
usage(const bool _ferr)
{
	const char * const s =
	    "Usage:\txmlmon [-v] -p pattern [-p pattern...] URL\n"
	    "\txmlmon -V | -h\n"
	    "\n"
	    "\t-h\tdisplay program usage information and exit\n"
	    "\t-p\tspecify a /document/first/second/node pattern to look for\n"
	    "\t-V\tdisplay program version information and exit\n"
	    "\t-v\tverbose operation; display diagnostic output\n";

	fprintf(_ferr? stderr: stdout, "%s", s);
	if (_ferr)
		exit(1);
}

static void
version(void)
{
	puts("xmlmon 0.1.0.dev1");
}

static void
debug(const char * const fmt, ...)
{
	va_list v;

	va_start(v, fmt);
	if (verbose)
		vfprintf(stderr, fmt, v);
	va_end(v);
}

static void __dead2
resp_exit(const char * const resp, const char * const fmt, va_list v)
{
	printf("%s ", resp);
	vprintf(fmt, v);
	putchar('\n');

	exit(0);
}

static void __dead2
resp_ok(const char * const fmt, ...)
{
	va_list v;
	va_start(v, fmt);
	resp_exit("OK", fmt, v);
	/* NOTREACHED */
}

static void __dead2
resp_err(const char * const fmt, ...)
{
	va_list v;
	va_start(v, fmt);
	resp_exit("ERR", fmt, v);
	/* NOTREACHED */
}

static void
handle_missing(void)
{
	resp_err("No patterns found in the XML document");
}

static void
handle_found(const char * const s)
{
	resp_ok("%s", s);
}

static char lcurl_errbuf[CURL_ERROR_SIZE];

static char xmlread;
static char xmlbuf[160];

static size_t
lcurl_read_data(void * const buf, size_t const membsz, size_t const nmemb, void * const cstate)
{
	struct mon_state * const state = cstate;
	const size_t sz = membsz * nmemb;
	debug("Processing %zu bytes of data", sz);
	axpath_parse_chunk(state->axstate, buf, sz);
	return (sz);
}

static void
complete_parsing(struct mon_state * const state __unused)
{
	axpath_parse_end(state->axstate);
}

static void
create_parser(struct mon_state * const state)
{
	axpath_create_parser(state->axstate, state->url, xmlbuf, xmlread);
	state->complete = complete_parsing;
}

static void
complete_create_parser(struct mon_state * const state)
{
	create_parser(state);
	complete_parsing(state);
}

static size_t
lcurl_create_parser(void * const buf, size_t const sz, void * const cstate)
{
	struct mon_state * const state = cstate;

	create_parser(state);

	if (curl_easy_setopt(state->curlh, CURLOPT_WRITEFUNCTION, lcurl_read_data) != CURLE_OK)
		resp_err("Could not switch to lcurl_read_data(): %s", lcurl_errbuf);
	if (sz == 0)
		return 0;
	return (lcurl_read_data(buf, 1, sz, cstate));
}

static size_t
lcurl_read_first_line(void * const buf, size_t const membsz, size_t const nmemb, void * const cstate)
{
	const size_t sz = membsz * nmemb;
	char * const cbuf = buf;
	const char * const nl_ptr = memchr(cbuf, '\n', sz);
	const size_t xmlleft = sizeof(xmlbuf) - xmlread;

	const size_t count =
		nl_ptr != NULL? (
			cbuf + xmlleft > nl_ptr + 1?
				(size_t)(nl_ptr - cbuf) + 1:
				xmlleft
			): (
			xmlleft > sz?
				sz:
				xmlleft
			);
	memcpy(xmlbuf + xmlread, cbuf, count);
	xmlread += count;

	if (sz > count || nl_ptr != NULL)
		return (count +
		    lcurl_create_parser(cbuf + count, sz - count, cstate));
	return (sz);
}

static CURL *
init_curl(struct mon_state * const state)
{
	const CURLcode rinit = curl_global_init(
	    CURL_GLOBAL_DEFAULT | CURL_GLOBAL_SSL);
	if (rinit != CURLE_OK)
		resp_err("Could not initialize the cURL library: %s",
		    curl_easy_strerror(rinit));

	CURL * const curlh = curl_easy_init();
	if (curlh == NULL)
		resp_err("Could not create a cURL handle");

	const CURLcode rset = curl_easy_setopt(curlh, CURLOPT_ERRORBUFFER,
	    lcurl_errbuf);
	if (rset != CURLE_OK)
		resp_err("Could not set the cURL handle's error buffer: %s",
		    curl_easy_strerror(rset));
	if (curl_easy_setopt(curlh, CURLOPT_WRITEFUNCTION, lcurl_read_first_line) != CURLE_OK)
		resp_err("Could not initialize the cURL read callback: %s",
		    lcurl_errbuf);
	if (curl_easy_setopt(curlh, CURLOPT_WRITEDATA, state) != CURLE_OK)
		resp_err("Could not initialize the cURL callback arg: %s",
		    lcurl_errbuf);
	return curlh;
}

int
main(int argc, char * const argv[])
{
	struct mon_state state;
	state.axstate = axpath_state_init(handle_found, handle_missing, resp_err);

	bool hflag = false, Vflag = false, have_patterns = false;
	int ch;
	while (ch = getopt(argc, argv, "hp:Vv"), ch != -1)
		switch (ch) {
			case 'h':
				hflag = true;
				break;

			case 'p':
				{
				axpath_add_pattern(state.axstate, optarg);
				have_patterns = true;
				}
				break;

			case 'V':
				Vflag = true;
				break;

			case 'v':
				verbose = true;
				break;

			default:
				usage(1);
				/* NOTREACHED */
		}
	if (Vflag)
		version();
	if (hflag)
		usage(false);
	if (Vflag || hflag)
		return (0);

	if (!have_patterns)
		resp_err("No pattern specified");

	argc -= optind;
	argv += optind;
	if (argc != 1)
		resp_err("Exactly one URL argument must be supplied");

	state.curlh = init_curl(&state);
	state.url = argv[0];
	debug("Passing an URL to libcurl: %s", state.url);
	if (curl_easy_setopt(state.curlh, CURLOPT_URL, state.url) != CURLE_OK)
		resp_err("Invalid URL '%s' for the cURL library: %s",
		    state.url, lcurl_errbuf);

	state.complete = complete_create_parser;
	if (curl_easy_perform(state.curlh) != CURLE_OK)
		resp_err("Could not fetch the data at '%s': %s",
		    state.url, lcurl_errbuf);
	state.complete(&state);
	resp_err("FIXME handle the cached stuff and stuff");
}
